# insert some parts here, find out later where they should go best
import pprint

from meresco.components import StorageComponent, XmlParseLxml, Lxml2Amara, XmlXPath, XmlPrintLxml, XmlParseAmara, XmlPrintAmara, Amara2Lxml, Venturi, Configuration, FilterField, RenameField, TransformFieldValue, RewritePartname, Xml2Fields, XPath2Field, FilterMessages, Rss, RssItem, Fields2XmlTx, Reindex, ReindexConsole, XsltCrosswalk
from meresco.components.statistics import Statistics, Logger
from meresco.components.statisticsxml import StatisticsXml
from meresco.components.drilldown import Drilldown, SRUDrilldownAdapter, SRUTermDrilldown, DrilldownRequestFieldnameMap
from meresco.components.http import PathFilter, FileServer, PathRename, IpFilter
from meresco.components.http.webrequestserver import WebRequestServer
from meresco.components.lucene import LuceneIndex, unlock, CQL2LuceneQuery, Fields2LuceneDocumentTx
from meresco.components.sru import Sru, SRURecordUpdate

from meresco.framework import Transparant, TransactionScope, ResourceManager, Observable

from cq2utils import DirectoryWatcher

from os.path import join, dirname, abspath, isdir
from os import remove, makedirs

# use homebrew dynamichtml from bzv instead of the cq2 version to support logging
from bzv.dynamichtml import DynamicHtml

from delfttilburg.normalize import LastnameNormalize, FirstInitialNormalize
from bzv.yearnormalize import YearNormalize
#from delfttilburg.normalize import YearNormalize, LastnameNormalize, FirstInitialNormalize
from delfttilburg import RateRecordCookie, RateRecord, MultiLevelDrilldown, ComposeNameFromMods, ModsDateField
from bzv.utils import HandleRequestPrinter
from bzv.openurl import OpenURL
from bzv.composedaifrommods import ComposeDaiFromMods
from bzv.record import Record, RecordXml, RecordJSON
from bzv.googlelikequery import GoogleLikeQuery
from bzv.xmldocumentadapter import XmlDocumentAdapter
from bzv.writedeleted import WriteDeleted
from bzv.undelete import Undelete
from bzv.getavailability import GetAvailability

unqualifiedTermFields = [('mods', 1.0), ('meta', 0.1)]

drilldownFieldnames = [
#    'drilldown.mods.name.namePart',
#    'drilldown.mods.name.type',
#    'drilldown.mods.name.role.roleTerm',
    'drilldown.mods.classification',
    'drilldown.mods.genre',
    'drilldown.mods.language.languageTerm',
    'drilldown.mods.relatedItem.titleInfo.title',
    'drilldown.mods.subject.topic',
    'drilldown.meta.repository.id',
    'drilldown.meta.repository.repositoryGroupId',
    'drilldown.norm.date.unparsable',
    'drilldown.norm.date.year',
    'drilldown.norm.date.decennium',
    'drilldown.norm.date.centennium',
    'drilldown.norm.name.oneinitial',
    'drilldown.norm.name.lastname',
    'drilldown.norm.name.unparsable',
]

#we gebruiken als sort fields drilldown fields en meta fields
sortFieldnames = []
#    'sort.mods.originInfo.dateIssued',
#    'sort.norm.date.year',
#]
metaFieldnames = [
    'meta.upload.id',
    'meta.record.id',
    'meta.repository.id',
    'meta.repository.set',
    'meta.repository.baseurl',
    'meta.repository.repositoryGroupId',
    'meta.repository.metadataPrefix',
    'meta.repository.collection',
]

DRILLDOWN_PREFIX = 'drilldown.'
SORT_PREFIX = 'sort.'

namespaces = {
    'document':'http://meresco.com/namespace/harvester/document',
    'oai_dc'  :'http://www.openarchives.org/OAI/2.0/oai_dc/',
    'meta'    :'http://meresco.com/namespace/harvester/meta',
    'mods'    :'http://www.loc.gov/mods/v3',
    'rating'  :'http://meresco.com/namespace/record/rating'
}

#xsltDir = join(dirname(abspath(__file__)), 'picaxslt')
#xsltFiles = [join(xsltDir, f) for f in ['picaplus2mods.xsl']]

class ServerLog(Observable, Logger):
    def handleRequest(self, Client=None, *args, **kwargs):
        if Client:
            self.log(ip=Client[0])
        self.log(request='request')
        return self.all.handleRequest(Client=Client, *args, **kwargs)

def createUploadHelix(index, storageComponent, xsltFiles ):
    yearNormalize = YearNormalize(yearRange=(1000,2100))
    firstInitialNormalize = FirstInitialNormalize()
    lastnameNormalize = LastnameNormalize()


    fields2LuceneDocument = (ResourceManager("record", lambda resourceManager: Fields2LuceneDocumentTx(resourceManager, untokenized=drilldownFieldnames + sortFieldnames + metaFieldnames + ['norm.dai'])),
        index
    )

    indexingHelix = \
        (Transparant(),
            fields2LuceneDocument,
            (FilterField(lambda name: DRILLDOWN_PREFIX + name in drilldownFieldnames),
                (RenameField(lambda name: DRILLDOWN_PREFIX + name),
                    fields2LuceneDocument
                )
            ),
            (FilterField(lambda name: SORT_PREFIX + name in sortFieldnames),
                (RenameField(lambda name: SORT_PREFIX + name),
                    fields2LuceneDocument
                )
            ),
        )

    normalizeTransactionStoreHelix = \
        (ResourceManager("record", lambda resourceManager:Fields2XmlTx(resourceManager, 'extra', namespace="http://meresco.com/namespace/fields/extra")),
            (storageComponent,)
        )

    storeModsHelix = \
        (RewritePartname('mods'),
            (XmlPrintLxml(),
                (storageComponent,)
            ),
	)

    indexModsHelix = \
        (XmlXPath(['//mods:mods'], namespaceMap={'mods': 'http://www.loc.gov/mods/v3'}),
            (Xml2Fields(),
                indexingHelix,
                (RenameField(lambda name: name.split('.')[0]),
                    indexingHelix
                ),
                (FilterField(lambda name: name == '__id__'),
                    normalizeTransactionStoreHelix
                )
            ),
            (ComposeNameFromMods('norm.name.oneinitial'),
                (TransformFieldValue(firstInitialNormalize.normalize),
                    indexingHelix,
                    normalizeTransactionStoreHelix,
                ),
                (TransformFieldValue(firstInitialNormalize.unparsable),
                    (RenameField(lambda name: 'norm.name.unparsable'),
                        indexingHelix
                    )
                ),
                (TransformFieldValue(lastnameNormalize.normalize),
                    (RenameField(lambda name: 'norm.name.lastname'),
                        indexingHelix,
                        normalizeTransactionStoreHelix,
                    )
                )
            ),
            (ComposeDaiFromMods('norm.dai'),
                indexingHelix,
                normalizeTransactionStoreHelix,
            ),
            (XPath2Field([('//mods:originInfo/mods:dateIssued/text()', 'norm.date.year')], namespaceMap={'mods': 'http://www.loc.gov/mods/v3'}),
                (TransformFieldValue(yearNormalize.normalize),
                    indexingHelix,
                    normalizeTransactionStoreHelix,
                    (TransformFieldValue(lambda year: year[:len('YYY')]+'0'),
                        (RenameField(lambda name: 'norm.date.decennium'),
                            indexingHelix,
                            normalizeTransactionStoreHelix
                        )
                    ),
                    (TransformFieldValue(lambda year: year[:len('YY')]+'00'),
                        (RenameField(lambda name: 'norm.date.centennium'),
                            indexingHelix,
                            normalizeTransactionStoreHelix
                        )
                    )
                ),
                (TransformFieldValue(yearNormalize.unparsable),
                    (RenameField(lambda name: 'norm.date.unparsable'),
                        indexingHelix
                    )
                )
            ),
            (ModsDateField(),
                indexingHelix,
            ),
        )

    return \
        (TransactionScope("record"),
            (Venturi(
                    should=[
                         ('metadata', '/document:document/document:part[@name="metadata"]/text()'),
                         ('meta', '/document:document/document:part[@name="meta"]/text()')],
                    could=[
                        ('rating', '/rating:rating'),
                        ('header', '/document:document/document:part[@name="header"]/text()'),],
                    namespaceMap= namespaces
                ),
                (FilterMessages(allowed=['delete']),
                    index,
                    (WriteDeleted(),
                    (storageComponent,))
                ),
                (FilterMessages(allowed=['getStream']),
                    (storageComponent,)
                ),

############### TP: storing of the required ('should') and optional ('could') parts
                (XmlPrintLxml(),
                    (storageComponent,)
                ),

                (XmlXPath(['//mods:mods'], namespaceMap={'mods': 'http://www.loc.gov/mods/v3'}),
		    indexModsHelix
                ),

                (XmlXPath(['/oai:metadata//record']),
                    (XsltCrosswalk(xsltFiles),
                        storeModsHelix,
			indexModsHelix
                    )
                ),

                (XmlXPath(['/meta:meta']),
                    (Xml2Fields(),
                        indexingHelix,
                        (FilterField(lambda name: name == 'meta.repository.id'),
                            (RenameField(lambda name: 'meresco.exists'),
                                (TransformFieldValue(lambda value: 'true'),
                                    indexingHelix
                                )
                            )
                        ),
                        (RenameField(lambda name: 'meta'),
                            indexingHelix
                        )
                    ),
                ),
                (FilterMessages(disallowed=['delete']),
                    (Undelete(),
                    (storageComponent,))
                )
            )
        )

def main(reactor, observableHttpServer, sruUpdateServer, inbox, host, portNumber, databasePath, \
         configDict, infra, configs):

    websiteConfig = configs ['website']
    serverConfig  = configs ['server']

    applicationName = websiteConfig.get('applicationName',section='userFeedback', default='Beter Zoeken en Vinden')
    storageDirectory = join(databasePath, 'storage')
    indexDirectory = join(databasePath, 'index')

    dynamicPath = configDict['dynamicPath']
    staticPath  = configDict['staticPath']
    xsltPath    = configDict['xsltPath']

    # space seperated filelist
    picaplusXsltFileList = configDict['picaplusXsltFiles'].split()
    xsltFiles = [join(xsltPath, f) for f in picaplusXsltFileList ]


    unlock(indexDirectory)

    storageComponent = StorageComponent(storageDirectory, partsRemovedOnDelete=[])
    drilldownComponent = Drilldown(drilldownFieldnames)

    indexHelix = \
            (LuceneIndex(indexDirectory,timer=reactor),
                (drilldownComponent,)
            )

    uploadHelix = createUploadHelix(indexHelix, storageComponent, xsltFiles )

    configurationComponent = Configuration(configDict)
    rateRecordList = RateRecordCookie(join(databasePath, 'ratedRecords'))

    statisticsPath = join(databasePath, 'statistics')
    isdir(statisticsPath) or makedirs(statisticsPath)
    statistics = Statistics(statisticsPath, [('ip',), ('clause',)], 60)

    locator = serverConfig.get('Locator',section='proxies')

    return \
        (Observable(),
            (inbox,
                (XmlDocumentAdapter(),
                    uploadHelix,
                )
            ),
            (sruUpdateServer,
                (PathFilter('/update'),
                    (WebRequestServer(),
                        (SRURecordUpdate(),
                            (Amara2Lxml(),
                                uploadHelix
                            )
                        )
                    )
                ),
                (PathFilter('/reindex'),
                    (IpFilter(['127.0.0.1']),
                        (ReindexConsole(),
                            (Reindex('meta'),
                                (FilterMessages(allowed=['listIdentifiers']),
                                    (storageComponent,)
                                ),
                                (FilterMessages(disallowed=['listIdentifiers']),
                                    uploadHelix
                                )
                            )
                        )
                    )
                )
            ),
            (observableHttpServer,
                (statistics,
                    (ServerLog(), ),
                    (PathFilter('/statistics'),
                        (StatisticsXml(statistics), )
                    ),
                    (PathFilter('/rss'),
                        (Rss(   title = applicationName,
                                description = 'Tilburg University',
                                link = host,
                                maximumRecords = 15),
                            (CQL2LuceneQuery(unqualifiedTermFields),
                                indexHelix
                            ),
                            (RssItem(
                                    nsMap=namespaces,
                                    title = ('mods', '//mods:mods/mods:titleInfo/mods:title/text()'),
                                    description = ('mods', '//mods:mods/mods:abstract/text()'),
                                    linkTemplate = host + '/page/recordview?recordId=%(uploadId)s',
                                    uploadId = ('meta', '/meta:meta/meta:upload/meta:id/text()')),
                                (storageComponent, )
                            ),
                        )
                    ),
		    (PathFilter('/record'),
		        (RecordXml(),
                            (Record(),
			        (storageComponent,)
                            )
                        )
                    ),
		    (PathFilter('/JSON'),
		        (RecordJSON(),
                            (Record(),
                                (storageComponent,)
                            )
                        )
                    ),
                    (PathFilter('/availability/'),
                        (GetAvailability(),)
                    ),
                    (PathFilter('/sru'),
                        (HandleRequestPrinter(),),
                        (Sru(host=host, port=portNumber, defaultRecordSchema='mods', defaultRecordPacking='xml'),
                            (CQL2LuceneQuery(unqualifiedTermFields),
                                indexHelix
                            ),
                            (storageComponent,),
                            (SRUDrilldownAdapter(),
                                (SRUTermDrilldown(),
                                    (DrilldownRequestFieldnameMap(
                                        lambda field: DRILLDOWN_PREFIX + field,
                                        lambda field: field[len(DRILLDOWN_PREFIX):]),
                                            (drilldownComponent,)
                                    ),
                                    (CQL2LuceneQuery(unqualifiedTermFields),
                                        indexHelix
                                    ),
                                )
                            )
                        )
                    ),
                    (PathFilter('/files'),\
                        (PathRename(lambda path: path[len('/files'):]),
                            (FileServer(staticPath),)
                        )
                    ),
                    (PathFilter('/', excluding=['/rss', '/record', '/JSON', '/availability/','/sru', '/update', '/files', '/statistics', '/reindex']),
                        (DynamicHtml([dynamicPath], reactor=reactor, allowedModules=['cqlparser', 'cqlwebutils'],
                                     indexPage='/nl/page/search',
                                     infra = infra,
                                     configs = configs
                                     ),
                            (GoogleLikeQuery(), ),
                            (CQL2LuceneQuery(unqualifiedTermFields),
                                indexHelix
                            ),
                            (OpenURL(locator),),
			    (Record(), (storageComponent,),),
                            (storageComponent,),
                            (drilldownComponent,),
                            (RateRecord(1,5),
                                (FilterMessages(allowed=['isAvailable', 'getStream']),
                                    (storageComponent,),
                                ),
                                uploadHelix
                            ),
                            (MultiLevelDrilldown({
                                    'date': [
                                        ('drilldown.norm.date.year', 20, False),
                                        ('drilldown.norm.date.decennium', 20, False),
                                        ('drilldown.norm.date.centennium', 10, False)
                                        ],
                                    'author': [
                                        ('drilldown.norm.name.oneinitial', 20, False),
                                        ('drilldown.norm.name.lastname', 10, True)
                                        ]
                                    }
                                ),
                                (drilldownComponent,),
                            ),
                            (configurationComponent,),
                            (rateRecordList, ),
                        )
                    )
                )
            )
        )
